@@ -9,11 +9,11 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
+(in-package "SB-IMPL")
 
 ;;;; miscellaneous non-primitive predicates
 
-#!-sb-fluid (declaim (inline streamp))
+#-sb-fluid (declaim (inline streamp))
 (defun streamp (stream)
   (typep stream 'stream))
 
@@ -26,7 +26,7 @@
                   collect `(defun ,name (x)
                              (or (typep x '(simple-array ,spec (*)))
                                  (and (complex-vector-p x)
-                                      (do ((data (%array-data-vector x) (%array-data-vector data)))
+                                      (do ((data (%array-data x) (%array-data data)))
                                           ((not (array-header-p data)) (typep data '(simple-array ,spec (*))))))))))))
   (def))
 
@@ -90,21 +90,24 @@
                     (stem (string-left-trim "%" (string-right-trim "P-" name)))
                     (article (if (position (schar name 0) "AEIOU") "an" "a")))
                `(defun ,pred (object)
-                  ,(format nil
-                           "Return true if OBJECT is ~A ~A, and NIL otherwise."
-                           article
-                           stem)
+                  ;; Document the standardized predicates and not the internal ones.
+                  ,@(when (eql (sb-xc:symbol-package pred) *cl-package*)
+                      (list (format nil
+                                    "Return true if OBJECT is ~A ~A, and NIL otherwise."
+                                    article
+                                    stem)))
                   ;; (falling through to low-level implementation)
                   (,pred object)))))
   (def-type-predicate-wrapper array-header-p)
+  (def-type-predicate-wrapper simple-array-header-p)
   (def-type-predicate-wrapper arrayp)
   (def-type-predicate-wrapper atom)
   ;; Testing for BASE-CHAR-P is usually redundant on #-sb-unicode,
   ;; remove it there completely so that #-sb-unicode build will
   ;; break when it's used.
-  #!+sb-unicode (def-type-predicate-wrapper base-char-p)
+  #+sb-unicode (def-type-predicate-wrapper base-char-p)
   (def-type-predicate-wrapper base-string-p)
-  #!+sb-unicode (def-type-predicate-wrapper character-string-p)
+  #+sb-unicode (def-type-predicate-wrapper character-string-p)
   (def-type-predicate-wrapper bignump)
   (def-type-predicate-wrapper bit-vector-p)
   (def-type-predicate-wrapper characterp)
@@ -114,7 +117,7 @@
   (def-type-predicate-wrapper complexp)
   (def-type-predicate-wrapper complex-double-float-p)
   (def-type-predicate-wrapper complex-float-p)
-  #!+long-float (def-type-predicate-wrapper complex-long-float-p)
+  #+long-float (def-type-predicate-wrapper complex-long-float-p)
   (def-type-predicate-wrapper complex-rational-p)
   (def-type-predicate-wrapper complex-single-float-p)
   ;; (COMPLEX-VECTOR-P is not included here since it's awkward to express
@@ -129,7 +132,7 @@
   (def-type-predicate-wrapper integerp)
   (def-type-predicate-wrapper listp)
   (def-type-predicate-wrapper long-float-p)
-  #!-(or x86 x86-64) (def-type-predicate-wrapper lra-p)
+  #-(or x86 x86-64) (def-type-predicate-wrapper lra-p)
   (def-type-predicate-wrapper null)
   (def-type-predicate-wrapper numberp)
   (def-type-predicate-wrapper rationalp)
@@ -137,17 +140,23 @@
   (def-type-predicate-wrapper realp)
   (def-type-predicate-wrapper short-float-p)
   (def-type-predicate-wrapper single-float-p)
-  #!+sb-simd-pack (def-type-predicate-wrapper simd-pack-p)
+  #+sb-simd-pack (def-type-predicate-wrapper simd-pack-p)
+  #+sb-simd-pack-256 (def-type-predicate-wrapper simd-pack-256-p)
   (def-type-predicate-wrapper %instancep)
+  (def-type-predicate-wrapper funcallable-instance-p)
   (def-type-predicate-wrapper symbolp)
+  ;; The interpreter needs this because it assumes that any type spec
+  ;; in SB-C::*BACKEND-TYPE-PREDICATES* has a callable predicate.
+  (def-type-predicate-wrapper non-null-symbol-p)
   (def-type-predicate-wrapper %other-pointer-p)
   (def-type-predicate-wrapper system-area-pointer-p)
+  (def-type-predicate-wrapper unbound-marker-p)
   (def-type-predicate-wrapper weak-pointer-p)
-  #!-64-bit
+  #-64-bit
   (progn
     (def-type-predicate-wrapper unsigned-byte-32-p)
     (def-type-predicate-wrapper signed-byte-32-p))
-  #!+64-bit
+  #+64-bit
   (progn
     (def-type-predicate-wrapper unsigned-byte-64-p)
     (def-type-predicate-wrapper signed-byte-64-p))
@@ -157,8 +166,8 @@
                   ,@(map 'list
                          (lambda (saetp)
                            `(def-type-predicate-wrapper
-                                ,(symbolicate (sb!vm:saetp-primitive-type-name saetp) "-P")))
-                         sb!vm:*specialized-array-element-type-properties*))))
+                                ,(symbolicate (sb-vm:saetp-primitive-type-name saetp) "-P")))
+                         sb-vm:*specialized-array-element-type-properties*))))
     (saetp-defs))
   ;; Other array types
   (def-type-predicate-wrapper simple-array-p)
@@ -168,25 +177,45 @@
   (def-type-predicate-wrapper vectorp)
   (def-type-predicate-wrapper vector-nil-p))
 
-#!+(or x86 x86-64 arm arm64)
+#+(or x86 x86-64 arm arm64)
 (defun fixnum-mod-p (x limit)
   (and (fixnump x)
        (<= 0 x limit)))
-
-#!+(or x86 x86-64 ppc)
-(defun %other-pointer-subtype-p (x choices)
-  (and (%other-pointer-p x)
-       (member (%other-pointer-widetag x) choices)
-       t))
 
+;;; a vector that maps widetags to layouts, used for quickly finding
+;;; the layouts of built-in classes
+(define-load-time-global **primitive-object-layouts** nil)
+(declaim (type simple-vector **primitive-object-layouts**))
+(defun !pred-cold-init ()
+  ;; This vector is allocated in immobile space when possible. There isn't
+  ;; a way to do that from lisp, so it's special-cased in genesis.
+  #-immobile-space (setq **primitive-object-layouts** (make-array 256))
+  (map-into **primitive-object-layouts**
+            (lambda (name) (classoid-layout (find-classoid name)))
+            #.(let ((table (make-array 256 :initial-element 'sb-kernel::random-class)))
+                (dolist (x sb-kernel::+!built-in-classes+)
+                  (destructuring-bind (name &key codes &allow-other-keys) x
+                    (dolist (code codes)
+                      (setf (svref table code) name))))
+                (loop for i from sb-vm:list-pointer-lowtag by (* 2 sb-vm:n-word-bytes)
+                      below 256
+                      do (setf (aref table i) 'cons))
+                (loop for i from sb-vm:even-fixnum-lowtag by (ash 1 sb-vm:n-fixnum-tag-bits)
+                      below 256
+                      do (setf (aref table i) 'fixnum))
+                table)))
+
 ;;; Return the layout for an object. This is the basic operation for
 ;;; finding out the "type" of an object, and is used for generic
 ;;; function dispatch. The standard doesn't seem to say as much as it
 ;;; should about what this returns for built-in objects. For example,
 ;;; it seems that we must return NULL rather than LIST when X is NIL
 ;;; so that GF's can specialize on NULL.
+;;; x86-64 has a vop that implements this without even needing to place
+;;; the vector of layouts in the constant pool of the containing code.
+#-(and compact-instance-header x86-64)
+(progn
 (declaim (inline layout-of))
-#-sb-xc-host
 (defun layout-of (x)
   (declare (optimize (speed 3) (safety 0)))
   (cond ((%instancep x) (%instance-layout x))
@@ -197,8 +226,8 @@
         (t
          ;; Note that WIDETAG-OF is slightly suboptimal here and could be
          ;; improved - we've already ruled out some of the lowtags.
-         (svref (load-time-value sb!kernel::**built-in-class-codes** t)
-                (widetag-of x)))))
+         (svref (load-time-value **primitive-object-layouts** t)
+                (widetag-of x))))))
 
 (declaim (inline classoid-of))
 #-sb-xc-host
@@ -222,10 +251,10 @@
      (cond
        ((<= 0 object 1) 'bit)
        ((< object 0) 'fixnum)
-       (t '(integer 0 #.sb!xc:most-positive-fixnum))))
+       (t `(integer 0 ,sb-xc:most-positive-fixnum))))
     (integer
      (if (>= object 0)
-         '(integer #.(1+ sb!xc:most-positive-fixnum))
+         `(integer ,(1+ sb-xc:most-positive-fixnum))
          'bignum))
     (character
      (typecase object
@@ -239,21 +268,21 @@
     (symbol
      (cond ((eq object t) 'boolean)
            ((eq object nil) 'null)
-           ((eq (symbol-package object) *keyword-package*) 'keyword)
+           ((eq (sb-xc:symbol-package object) *keyword-package*) 'keyword)
            (t 'symbol)))
-    ((or array complex #!+sb-simd-pack simd-pack)
-     (let ((sb!kernel::*unparse-allow-negation* nil))
-       (declare (special sb!kernel::*unparse-allow-negation*)) ; forward ref
+    ((or array complex #+sb-simd-pack simd-pack #+sb-simd-pack-256 simd-pack-256)
+     (let ((sb-kernel::*unparse-allow-negation* nil))
+       (declare (special sb-kernel::*unparse-allow-negation*)) ; forward ref
        (type-specifier (ctype-of object))))
     (t
      (let* ((classoid (classoid-of object))
             (name (classoid-name classoid)))
        (if (%instancep object)
            (case name
-             (sb!alien-internals:alien-value
+             (sb-alien-internals:alien-value
               `(alien
-                ,(sb!alien-internals:unparse-alien-type
-                  (sb!alien-internals:alien-value-type object))))
+                ,(sb-alien-internals:unparse-alien-type
+                  (sb-alien-internals:alien-value-type object))))
              (t
               (let ((pname (classoid-proper-name classoid)))
                 (if (classoid-p pname)
@@ -269,22 +298,28 @@
   (eq obj1 obj2))
 ;;; and this too, but it's only needed for backends on which
 ;;; IR1 might potentially transform EQL into %EQL/INTEGER.
-#!+integer-eql-vop
+#+integer-eql-vop
 (defun %eql/integer (obj1 obj2)
-  ;; This is just for constant folding, no need to transform into the %EQL/INTEGER VOP
-  (eql obj1 obj2))
+  ;; This is just for constant folding, there's no real need to transform
+  ;; into the %EQL/INTEGER VOP. But it's confusing if it becomes identical to
+  ;; EQL, and then due to ICF we find that #'EQL => #<FUNCTION %EQL/INTEGER>.
+  ;; Type declarations don't suffice because we don't know which arg is an integer
+  ;; (if not both). We could ensure selection of the vop by using %PRIMITIVE.
+  ;; Anyway it suffices to disable type checking and pretend its the always
+  ;; the first arg that's an integer, but that won't work on the host because
+  ;; it might enforce the type since we can't portably unenforce after declaring.
+  #-sb-xc-host (declare (optimize (sb-c::type-check 0)))
+  (eql #+sb-xc-host obj1 #-sb-xc-host (the integer obj1) obj2))
 
 (declaim (inline %eql))
 (defun %eql (obj1 obj2)
   "Return T if OBJ1 and OBJ2 represent the same object, otherwise NIL."
+  #+x86-64 (eql obj1 obj2) ; vop fully implements all cases of EQL
+  #-x86-64 ; else this is the only full implementation of EQL
   (or (eq obj1 obj2)
       (if (or (typep obj2 'fixnum)
               (not (typep obj2 'number)))
           nil
-          ;; I would think that we could do slightly better here by testing that
-          ;; both objs are OTHER-POINTER-P with equal %OTHER-POINTER-WIDETAGs.
-          ;; Then dispatch on obj2 and elide the TYPEP on obj1 using TRULY-THE.
-          ;; Also would need to deal with immediate single-float for 64-bit.
           (macrolet ((foo (&rest stuff)
                        `(typecase obj2
                           ,@(mapcar (lambda (foo)
@@ -297,11 +332,11 @@
             (foo
              (single-float eql)
              (double-float eql)
-             #!+long-float
+             #+long-float
              (long-float eql)
              (bignum
-              #!-integer-eql-vop (lambda (x y) (zerop (bignum-compare x y)))
-              #!+integer-eql-vop eql) ; will become %eql/integer
+              #-integer-eql-vop (lambda (x y) (zerop (bignum-compare x y)))
+              #+integer-eql-vop eql) ; will become %eql/integer
              (ratio
               (lambda (x y)
                 (and (eql (numerator x) (numerator y))
@@ -320,7 +355,9 @@
                      (eql (imagpart x) (imagpart y))))))))))
 
 (defun eql (x y)
-  (%eql x y))
+  ;; On x86-64, EQL is just an interpreter stub for a vop.
+  ;; For others it's a call to the implementation of generic EQL.
+  (#+x86-64 eql #-x86-64 %eql x y))
 
 (defun bit-vector-= (x y)
   (declare (type bit-vector x y))
@@ -367,6 +404,18 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
     (declare (maybe-inline equal-aux))
     (equal-aux x y)))
 
+;;; Like EQUAL, but any two gensyms whose names are STRING= are equalish.
+(defun fun-names-equalish (x y)
+  (named-let recurse ((x x) (y y))
+    (cond ((eql x y) t) ; not performance-critical: don't inline %EQL here
+          ((consp x) (and (consp y)
+                          (recurse (car x) (car y))
+                          (recurse (cdr x) (cdr y))))
+          ((and (symbolp x) (not (sb-xc:symbol-package x)))
+           (and (symbolp y) (not (sb-xc:symbol-package y)) (string= x y)))
+          (t
+           (equal x y)))))
+
 ;;; EQUALP comparison of HASH-TABLE values
 (defun hash-table-equalp (x y)
   (declare (type hash-table x y))
@@ -383,38 +432,34 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
                       x)
              t))))
 
+(declaim (inline instance-equalp))
 (defun instance-equalp (x y)
   (let ((layout-x (%instance-layout x)))
     (and
      (eq layout-x (%instance-layout y))
-     ;; TODO: store one bit in the layout indicating whether EQUALP
-     ;; should scan slots. (basically a STRUCTURE-CLASSOID-P bit)
-     (structure-classoid-p (layout-classoid layout-x))
+     (logtest +structure-layout-flag+ (layout-%bits layout-x))
      (macrolet ((slot-ref-equalp ()
                   `(let ((x-el (%instance-ref x i))
                          (y-el (%instance-ref y i)))
                      (or (eq x-el y-el) (equalp x-el y-el)))))
-         (if (eql (layout-bitmap layout-x) sb!kernel::+layout-all-tagged+)
-             (loop for i of-type index from sb!vm:instance-data-start
-                   below (layout-length layout-x)
+       (let ((n (%instance-length x)))
+         (if (eql (layout-bitmap layout-x) sb-kernel::+layout-all-tagged+)
+             (loop for i downfrom (1- n) to sb-vm:instance-data-start
                    always (slot-ref-equalp))
              (let ((comparators (layout-equalp-tests layout-x)))
-               (unless (= (length comparators)
-                          (- (layout-length layout-x) sb!vm:instance-data-start))
+               (unless (= (length comparators) (- n sb-vm:instance-data-start))
                  (bug "EQUALP got incomplete instance layout"))
                ;; See remark at the source code for %TARGET-DEFSTRUCT
                ;; explaining how to use the vector of comparators.
-               (loop for i of-type index from sb!vm:instance-data-start
-                     below (layout-length layout-x)
+               (loop for i downfrom (1- n) to sb-vm:instance-data-start
                      for test = (data-vector-ref
-                                 comparators (- i sb!vm:instance-data-start))
+                                 comparators (- i sb-vm:instance-data-start))
                      always (cond ((eql test 0) (slot-ref-equalp))
-                                  ((functionp test)
-                                   (funcall test i x y))
-                                  (t)))))))))
+                                  ((functionp test) (funcall test i x y))
+                                  (t))))))))))
 
 ;;; Doesn't work on simple vectors
-(defun array-equal-p (x y)
+(defun array-equalp (x y)
   (declare (array x y))
   (let ((rank (array-rank x)))
     (and
@@ -480,34 +525,39 @@ length and have identical components. Other arrays must be EQ to be EQUAL."
         ((%instancep x)
          (and (%instancep y)
               (instance-equalp x y)))
-        ((and (bit-vector-p x)
-              (bit-vector-p y))
+        ((and (simple-vector-p x) (simple-vector-p y))
+         (let ((len (length x)))
+           (and (= len (length y))
+                (loop for i below len ; somewhat faster than the generic loop
+                      always (let ((a (svref x i)) (b (svref y i)))
+                               (or (eq a b) (equalp a b)))))))
+        ((and (bit-vector-p x) (bit-vector-p y))
          (bit-vector-= x y))
         ((vectorp x)
          (and (vectorp y)
               (vector-equalp x y)))
         ((arrayp x)
          (and (arrayp y)
-              (array-equal-p x y)))
+              (array-equalp x y)))
         (t nil)))
 
-(/show0 "about to do test cases in pred.lisp")
-(let ((test-cases `((0.0 ,(load-time-value (make-unportable-float :single-float-negative-zero)) t)
-                    (0.0 1.0 nil)
-                    (#c(1 0) #c(1.0 0.0) t)
-                    (#c(0 1) #c(0.0 1.0) t)
-                    (#c(1.1 0.0) #c(11/10 0) nil) ; due to roundoff error
+(let ((test-cases `(($0.0 $-0.0 t)
+                    ($0.0 $1.0 nil)
+                    ;; There is no cross-compiler #C reader macro.
+                    ;; SB-XC:COMPLEX does not want uncanonical input, i.e. imagpart
+                    ;; of rational 0 which downgrades the result to just an integer.
+                    (1 ,(complex $1.0 $0.0) t)
+                    (,(complex 0 1) ,(complex $0.0 $1.0) t)
+                    ;; 11/10 is unequal to real 1.1 due to roundoff error.
+                    ;; COMPLEX here is a red herring
+                    (,(complex $1.1 $0.0) 11/10 nil)
                     ("Hello" "hello" t)
                     ("Hello" #(#\h #\E #\l #\l #\o) t)
                     ("Hello" "goodbye" nil))))
-  (/show0 "TEST-CASES bound in pred.lisp")
   (dolist (test-case test-cases)
-    (/show0 "about to do a TEST-CASE in pred.lisp")
     (destructuring-bind (x y expected-result) test-case
       (let* ((result (equalp x y))
              (bresult (if result 1 0))
              (expected-bresult (if expected-result 1 0)))
         (unless (= bresult expected-bresult)
-          (/show0 "failing test in pred.lisp")
           (error "failed test (EQUALP ~S ~S)" x y))))))
-(/show0 "done with test cases in pred.lisp")
