@@ -82,7 +82,8 @@
 ;;; but there's really no easy way to do that at present.
 (defconstant dchunk-bits
   #+x86-64 56
-  #-x86-64 sb-vm:n-word-bits)
+  #+ppc64 32
+  #-(or x86-64 ppc64) sb-vm:n-word-bits)
 
 (deftype dchunk ()
   `(unsigned-byte ,dchunk-bits))
