@@ -31,8 +31,8 @@
 
 (declaim (inline make-wrapper-internal))
 (defun make-wrapper-internal (&key (bitmap -1) length classoid)
-  (make-layout :length length :classoid classoid :invalid nil
-               :%for-std-class-b 1 :bitmap bitmap))
+  (make-layout classoid :length length :invalid nil
+               :flags +pcl-object-layout-flag+ :bitmap bitmap))
 
 ;;; With compact-instance-header and immobile-code, the primitive object has
 ;;; 2 descriptor slots (fin-fun and CLOS slot vector) and 2 non-desriptor slots
@@ -141,12 +141,59 @@
 (defun invalid-wrapper-p (wrapper)
   (not (null (layout-invalid wrapper))))
 
-;;; We only use this inside INVALIDATE-WRAPPER.
-(defvar *previous-nwrappers* (make-hash-table :test #'eq))
-
+;;; The portable code inherited from the ancient PCL sources used an extremely
+;;; clever mechanism to store a mapping from new layouts ("wrappers") to old.
+;;; Old layouts point to new in a straighforward way - via the INVALID slot -
+;;; but the invalidation logic wants backpointers from new to old as well.
+;;; It arranges so that if the following state exists:
+;;;   layout_A -> (:flush layout_B)
+;;; and then we want to change layout_B from valid to invalid:
+;;;   layout_B -> (:flush layout_C)
+;;; the state of layout_A should be changed to (:flush layout_C)
+;;; at the same time that the state of layout_B is changed.
+;;;
+;;; In general, alterating layout_B might might have to perform more than one
+;;; update to older versions of the layout, so it's a one-to-many relation.
+;;; The same "transitivity" could have been implemented at the time of flushing
+;;; layout_A by chasing an extra pointer (the B -> C pointer) and recording that
+;;; new state back into A. So I don't see how this "clever" code actually
+;;; improves anything, as it merely performs work eagerly that could have
+;;; been performed lazily. (PCL generally prefers lazy updates.)
+;;;
+;;; Anyway the extreme cleverness (and premature optimization, imho), was almost,
+;;; but not quite, clever enough - it leaked three conses per redefinition,
+;;; which it pretty much had to in order to remain portable.
+;;; We can at least replicate it in a way that doesn't leak conses,
+;;; or at least theoretically doesn't. A likely scenario is that old caches may
+;;; never get adquately flushed to drop all references to obsolete layouts.
+;;; If you call a GF once and not again, it won't flush; thus causing a retained
+;;; old layout no matter what else is done.
+;;;
+;;; So the non-leaky thing to do is use weak pointers, which makes the object
+;;; representation more transparent, because you can examine a classoid to get
+;;; the mapping rather than having to guess at what layouts are pointed to
+;;; by the variable formerly known as *previous-nwrappers*.
+;;;
+;;; And the code no longer relies on such spooky action at a distance - smashing
+;;; the CAR and CDR of the cons cells that are shared across layouts - but instead
+;;; performs a slot update on old layouts which, while still being a mutating
+;;; operation, is far more obvious as to intent.
+;;; This might even fix a data race, though I can't be certain.
+;;; The bug is that it was possible for an observer to see (:FLUSH new-thing)
+;;; in an older layout when it should actually have seen (:OBSOLETE newer-thing).
+;;; On the other hand, it's still possible for an observer to see the
+;;; wrong value in the INVALID slot itself. i.e. if it reads the slot
+;;; before %invalidate-wrapper performs the transitivity operation,
+;;; hence the possibility that this optimization is not only premature,
+;;; but also pretty much not correct.
 (defun %invalidate-wrapper (owrapper state nwrapper)
-  (aver (member state '(:flush :obsolete) :test #'eq))
-  (let ((new-previous ()))
+  (aver (member state '(:flush :obsolete)))
+  #+sb-thread (aver (eq (sb-thread:mutex-owner sb-c::**world-lock**)
+                        sb-thread:*current-thread*))
+  (let ((classoid (layout-classoid nwrapper))
+        (new-previous ())
+        (new-state (cons state nwrapper)))
+    (aver (eq (layout-classoid owrapper) classoid))
     ;; First off, a previous call to INVALIDATE-WRAPPER may have
     ;; recorded OWRAPPER as an NWRAPPER to update to. Since OWRAPPER
     ;; is about to be invalid, it no longer makes sense to update to
@@ -155,23 +202,30 @@
     ;; We go back and change the previously invalidated wrappers so
     ;; that they will now update directly to NWRAPPER. This
     ;; corresponds to a kind of transitivity of wrapper updates.
-    (dolist (previous (gethash owrapper *previous-nwrappers*))
-      (when (eq state :obsolete)
-        (setf (car previous) :obsolete))
-      (setf (cadr previous) nwrapper)
-      (push previous new-previous))
+    (dolist (weak-pointer (sb-kernel::standard-classoid-old-layouts classoid))
+      (let ((previous (weak-pointer-value weak-pointer)))
+        (when previous
+          (setf (layout-invalid previous)
+                (cond ((and (eq (car (layout-invalid previous)) :obsolete)
+                            (eq state :flush))
+                       ;; :obsolete must stay :obsolete,
+                       ;; requiring the protocol for obsolete instances.
+                       (cons :obsolete nwrapper))
+                      (t
+                       ;; flush->{obsolete|flush} or {flush|obsolete}->obsolete
+                       new-state)))
+          (push weak-pointer new-previous))))
 
     ;; FIXME: We are here inside PCL lock, but might someone be
     ;; accessing the wrapper at the same time from outside the lock?
     (setf (layout-clos-hash owrapper) 0)
-
-    ;; FIXME: We could save a whopping cons by using (STATE . WRAPPER)
-    ;; instead
-    (push (setf (layout-invalid owrapper) (list state nwrapper))
-          new-previous)
-
-    (remhash owrapper *previous-nwrappers*)
-    (setf (gethash nwrapper *previous-nwrappers*) new-previous)))
+    ;; And shouldn't these assignments be flipped, so that if an observer
+    ;; sees a 0 hash in owrapper, it can find the new layout?
+    (setf (layout-invalid owrapper) new-state)
+    (push (make-weak-pointer owrapper) new-previous)
+    ;; This function is called for effect; return value is arbitrary.
+    (setf  (sb-kernel::standard-classoid-old-layouts classoid)
+           new-previous)))
 
 ;;; FIXME: This is not a good name: part of the contract here is that
 ;;; we return the valid wrapper, which is not obvious from the name
@@ -223,7 +277,7 @@
             ((consp state)
              (ecase (car state)
                (:flush
-                (let ((new (cadr state)))
+                (let ((new (cdr state)))
                   (cond ((std-instance-p instance)
                          (setf (%instance-layout instance) new))
                         ((fsc-instance-p instance)
@@ -231,7 +285,7 @@
                         (t
                          (bug "unrecognized instance type")))))
                (:obsolete
-                (%obsolete-instance-trap owrapper (cadr state) instance))))))))
+                (%obsolete-instance-trap owrapper (cdr state) instance))))))))
 
 (declaim (inline check-obsolete-instance))
 (defun check-obsolete-instance (instance)
