@@ -49,12 +49,7 @@
 
 #include "validate.h"
 #include "thread.h"
-#include "gc.h"
-#if defined LISP_FEATURE_GENCGC
-#include "gencgc-internal.h"
-#else
-#include "cheneygc-internal.h"
-#endif
+#include "gc-internal.h"
 #include <fcntl.h>
 #ifdef LISP_FEATURE_SB_WTIMER
 # include <sys/timerfd.h>
@@ -127,6 +122,25 @@ futex_init()
     }
 }
 
+/* Try to guess the name of the mutex for this futex, based on knowing
+ * the two pertinent Lisp object types (WAITQUEUE and MUTEX) that use a futex.
+ * Callable from a C debugger */
+#include "genesis/vector.h"
+char* futex_name(int *lock_word)
+{
+    // If there is a Lisp string at lock_word-1 or -2, return that.
+    // Otherwise return NULL.
+    lispobj name = *(lock_word - 1);
+    struct vector* v = (struct vector*)native_pointer(name);
+    if (lowtag_of(name) == OTHER_POINTER_LOWTAG && widetag_of(&v->header) == SIMPLE_BASE_STRING_WIDETAG)
+        return (char*)(v->data);
+    name = *(lock_word - 2);
+    v = (struct vector*)native_pointer(name);
+    if (lowtag_of(name) == OTHER_POINTER_LOWTAG && widetag_of(&v->header) == SIMPLE_BASE_STRING_WIDETAG)
+        return (char*)(v->data);
+    return 0;
+}
+
 int
 futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
 {
@@ -178,29 +192,31 @@ isnptl (void)
 }
 #endif
 
-void
-os_init(char *argv[], char *envp[])
+static void getuname(int *major_version, int* minor_version, int *patch_version)
 {
     /* Conduct various version checks: do we have enough mmap(), is
      * this a sparc running 2.2, can we do threads? */
     struct utsname name;
-    int major_version;
-    int minor_version;
-    int patch_version;
     char *p;
     uname(&name);
 
     p=name.release;
-    major_version = atoi(p);
-    minor_version = patch_version = 0;
+    *major_version = atoi(p);
+    *minor_version = *patch_version = 0;
     p=strchr(p,'.');
     if (p != NULL) {
-            minor_version = atoi(++p);
+            *minor_version = atoi(++p);
             p=strchr(p,'.');
             if (p != NULL)
-                    patch_version = atoi(++p);
+                    *patch_version = atoi(++p);
     }
+}
 
+void os_init(char __attribute__((unused)) *argv[],
+             char __attribute__((unused)) *envp[])
+{
+    int major_version, minor_version, patch_version;
+    getuname(&major_version, &minor_version, &patch_version);
     if (major_version<2) {
         lose("linux kernel version too old: major version=%d (can't run in version < 2.0.0)\n",
              major_version);
@@ -228,15 +244,48 @@ os_init(char *argv[], char *envp[])
      */
     os_vm_page_size = BACKEND_PAGE_BYTES;
 
+#ifdef LISP_FEATURE_X86
+    /* Use SSE detector.  Recent versions of Linux enable SSE support
+     * on SSE capable CPUs.  */
+    /* FIXME: Are there any old versions that does not support SSE?  */
+    fast_bzero_pointer = fast_bzero_detect;
+#endif
+}
+
+#if (defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)) \
+     && (!defined(DISABLE_ASLR) || DISABLE_ASLR)
+# define ALLOW_PERSONALITY_CHANGE 1
+#else
+# define ALLOW_PERSONALITY_CHANGE 0
+#endif
+
+int os_preinit(char *argv[], char *envp[])
+{
+#if ALLOW_PERSONALITY_CHANGE
+    if (getenv("SBCL_IS_RESTARTING")) {
+        /* We restarted due to previously enabled ASLR.  Now,
+         * reenable it for fork()'ed children. */
+        int pers = personality(0xffffffffUL);
+        personality(pers & ~ADDR_NO_RANDOMIZE);
+        unsetenv("SBCL_IS_RESTARTING");
+        return 0; // ensure_spaces() will win or not. Not much to do here.
+    }
+#endif
+    /* See if we can allocate read-only, static, and linkage table spaces
+     * at their required addresses. If we can, then there's no need to try
+     * the re-exec trick since dynamic space is relocatable. */
+    if (allocate_hardwired_spaces(0)) // soft failure mode
+        return 1; // indicate that we already allocated hardwired spaces
+
     /* KLUDGE: Disable memory randomization on new Linux kernels
      * by setting a personality flag and re-executing. (We need
      * to re-execute, since the memory maps that can conflict with
      * the SBCL spaces have already been done at this point).
-     *
-     * Since randomization is currently implemented only on x86 kernels,
-     * don't do this trick on other platforms.
      */
-#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
+    int major_version, minor_version, patch_version;
+    getuname(&major_version, &minor_version, &patch_version);
+
+#if ALLOW_PERSONALITY_CHANGE
     if ((major_version == 2
          /* Some old kernels will apparently lose unsupported personality flags
           * on exec() */
@@ -276,29 +325,12 @@ os_init(char *argv[], char *envp[])
                     execv(runtime, argv);
                 }
             }
-            /* Either changing the personality or execve() failed. Either
-             * way we might as well continue, and hope that the random
-             * memory maps are ok this time around.
-             */
-            fprintf(stderr, "WARNING:\
-\nCouldn't re-execute SBCL with proper personality flags (/proc isn't mounted? setuid?)\
-\nTrying to continue anyway.\n");
-        } else if (getenv("SBCL_IS_RESTARTING")) {
-            /* We restarted due to previously enabled ASLR.  Now,
-             * reenable it for fork()'ed children. */
-            int pers = personality(0xffffffffUL);
-            personality(pers & ~ADDR_NO_RANDOMIZE);
-
-            unsetenv("SBCL_IS_RESTARTING");
+            /* Either changing the personality or execve() failed.
+             * Just get on with life and hope for the best. */
         }
     }
-#ifdef LISP_FEATURE_X86
-    /* Use SSE detector.  Recent versions of Linux enable SSE support
-     * on SSE capable CPUs.  */
-    /* FIXME: Are there any old versions that does not support SSE?  */
-    fast_bzero_pointer = fast_bzero_detect;
-#endif
 #endif
+    return 0;
 }
 
 
@@ -308,53 +340,20 @@ os_init(char *argv[], char *envp[])
  * information loss, we have to make sure it allocates all its ram in the
  * 0-2Gb region.  */
 
-static void * under_2gb_free_pointer=DYNAMIC_1_SPACE_END;
-#endif
-
-os_vm_address_t
-os_validate(os_vm_address_t addr, os_vm_size_t len)
+static void * under_2gb_free_pointer;
+os_set_cheneygc_spaces(uword_t space0_start, uword_t space1_start)
 {
-    int flags =  MAP_PRIVATE | MAP_ANONYMOUS | MAP_NORESERVE;
-    os_vm_address_t actual;
-
-#ifdef LISP_FEATURE_ALPHA
-    if (!addr) {
-        addr=under_2gb_free_pointer;
-    }
-#endif
-    actual = mmap(addr, len, OS_VM_PROT_ALL, flags, -1, 0);
-    if (actual == MAP_FAILED) {
-        perror("mmap");
-        return 0;               /* caller should check this */
-    }
-
-    if (addr && (addr!=actual)) {
-        fprintf(stderr, "mmap: wanted %lu bytes at %p, actually mapped at %p\n",
-                (unsigned long) len, addr, actual);
-        return 0;
-    }
-
-#ifdef LISP_FEATURE_ALPHA
-
-    len=(len+(os_vm_page_size-1))&(~(os_vm_page_size-1));
-    under_2gb_free_pointer+=len;
-#endif
-
-    return actual;
+    uword_t max;
+    max = (space1_start > space0_start) ? space1_start : space0_start;
+    under_2gb_free_pointer = max + dynamic_space_size;
 }
 
-void
-os_invalidate(os_vm_address_t addr, os_vm_size_t len)
-{
-    if (munmap(addr,len) == -1) {
-        perror("munmap");
-    }
-}
+#endif
 
 void
 os_protect(os_vm_address_t address, os_vm_size_t length, os_vm_prot_t prot)
 {
-    if (mprotect(address, length, prot) == -1) {
+    if (mprotect(address, length, prot)) {
         if (errno == ENOMEM) {
             lose("An mprotect call failed with ENOMEM. This probably means that the maximum amount\n"
                  "of separate memory mappings was exceeded. To fix the problem, either increase\n"
@@ -422,8 +421,10 @@ sigsegv_handler(int signal, siginfo_t *info, os_context_t *context)
 void
 os_install_interrupt_handlers(void)
 {
+    if (INSTALL_SIG_MEMORY_FAULT_HANDLER) {
     undoably_install_low_level_interrupt_handler(SIG_MEMORY_FAULT,
                                                  sigsegv_handler);
+    }
 
     /* OAOOM c.f. sunos-os.c.
      * Should we have a reusable function gc_install_interrupt_handlers? */
@@ -440,7 +441,7 @@ os_install_interrupt_handlers(void)
 }
 
 char *
-os_get_runtime_executable_path(int external)
+os_get_runtime_executable_path(int __attribute__((unused)) external)
 {
     char path[PATH_MAX + 1];
     int size;
