@@ -69,9 +69,7 @@ os_zero(os_vm_address_t addr, os_vm_size_t length)
         addr = os_validate(NOT_MOVABLE, block_start, block_size);
 
         if (addr == NULL || addr != block_start)
-            lose("os_zero: block moved! 0x%08x ==> 0x%08x\n",
-                 block_start,
-                 addr);
+            lose("os_zero: block moved! %p ==> %p\n", block_start, addr);
     }
 }
 #endif
@@ -148,7 +146,6 @@ void *
 os_dlsym_default(char *name)
 {
     void *frob = dlsym(RTLD_DEFAULT, name);
-    odxprint(misc, "%p", frob);
     return frob;
 }
 #endif
@@ -163,21 +160,20 @@ void os_link_runtime()
     char *namechars;
     boolean datap;
     void* result;
-    int n = 0, m = 0, j;
+    int j;
 
     if (lisp_linkage_table_n_prelinked)
         return; // Linkage was already performed by coreparse
 
     struct vector* symbols = VECTOR(SymbolValue(REQUIRED_FOREIGN_SYMBOLS,0));
-    n = lisp_linkage_table_n_prelinked = fixnum_value(symbols->length);
-    for (j = 0 ; j < n ; ++j)
+    lisp_linkage_table_n_prelinked = fixnum_value(symbols->length);
+    for (j = 0 ; j < lisp_linkage_table_n_prelinked ; ++j)
     {
         lispobj item = symbols->data[j];
         datap = listp(item);
         symbol_name = datap ? CONS(item)->car : item;
         namechars = (void*)(intptr_t)(VECTOR(symbol_name)->data);
         result = os_dlsym_default(namechars);
-        odxprint(runtime_link, "linking %s => %p", namechars, result);
 
         if (link_target == validated_end) {
             validated_end = (char*)validated_end + os_vm_page_size;
@@ -187,17 +183,35 @@ void os_link_runtime()
         }
         if (result) {
             arch_write_linkage_table_entry(link_target, result, datap);
-        } else {
-            m++;
+        } else { // startup might or might not work. ymmv
+            printf("Missing required foreign symbol '%s'\n", namechars);
         }
 
         link_target += LINKAGE_TABLE_ENTRY_SIZE;
     }
-    odxprint(runtime_link, "%d total symbols linked, %d undefined",
-             n, m);
 #endif /* LISP_FEATURE_SB_DYNAMIC_CORE */
 }
 
+boolean
+gc_managed_heap_space_p(lispobj addr)
+{
+    if ((READ_ONLY_SPACE_START <= addr && addr < READ_ONLY_SPACE_END)
+        || (STATIC_SPACE_START <= addr && addr < STATIC_SPACE_END)
+#if defined LISP_FEATURE_GENCGC
+        || (DYNAMIC_SPACE_START <= addr &&
+            addr < (DYNAMIC_SPACE_START + dynamic_space_size))
+        || immobile_space_p(addr)
+#else
+        || (DYNAMIC_0_SPACE_START <= addr &&
+            addr < DYNAMIC_0_SPACE_START + dynamic_space_size)
+        || (DYNAMIC_1_SPACE_START <= addr &&
+            addr < DYNAMIC_1_SPACE_START + dynamic_space_size)
+#endif
+        )
+        return 1;
+    return 0;
+}
+
 #ifndef LISP_FEATURE_WIN32
 
 /* Remap a part of an already existing mapping to a file */
@@ -209,35 +223,24 @@ void os_map(int fd, int offset, os_vm_address_t addr, os_vm_size_t len)
                   fd, (off_t) offset);
     if (actual == MAP_FAILED || (addr && (addr != actual))) {
         perror("mmap");
-        lose("unexpected mmap(%d, %d, ...) failure\n", addr, len);
+        lose("unexpected mmap(%"OBJ_FMTX", %"OBJ_FMTX") failure\n",
+             (lispobj)addr, (lispobj)len);
     }
 }
 
 boolean
-gc_managed_addr_p(lispobj ad)
+gc_managed_addr_p(lispobj addr)
 {
     struct thread *th;
 
-    if ((READ_ONLY_SPACE_START <= ad && ad < READ_ONLY_SPACE_END)
-        || (STATIC_SPACE_START <= ad && ad < STATIC_SPACE_END)
-#if defined LISP_FEATURE_GENCGC
-        || (DYNAMIC_SPACE_START <= ad &&
-            ad < (DYNAMIC_SPACE_START + dynamic_space_size))
-        || immobile_space_p(ad)
-#else
-        || (DYNAMIC_0_SPACE_START <= ad &&
-            ad < DYNAMIC_0_SPACE_START + dynamic_space_size)
-        || (DYNAMIC_1_SPACE_START <= ad &&
-            ad < DYNAMIC_1_SPACE_START + dynamic_space_size)
-#endif
-        )
+    if (gc_managed_heap_space_p(addr))
         return 1;
     for_each_thread(th) {
-        if(th->control_stack_start <= (lispobj*)ad
-           && (lispobj*)ad < th->control_stack_end)
+        if(th->control_stack_start <= (lispobj*)addr
+           && (lispobj*)addr < th->control_stack_end)
             return 1;
-        if(th->binding_stack_start <= (lispobj*)ad
-           && (lispobj*)ad < th->binding_stack_start + BINDING_STACK_SIZE)
+        if(th->binding_stack_start <= (lispobj*)addr
+           && (lispobj*)addr < th->binding_stack_start + BINDING_STACK_SIZE)
             return 1;
     }
     return 0;
